package info.textgrid.utils.export.filenames;

import info.textgrid.namespaces.metadata.agent._2010.AgentRoleType;
import info.textgrid.namespaces.metadata.agent._2010.AgentType;
import info.textgrid.namespaces.metadata.core._2010.ObjectType;
import info.textgrid.utils.export.aggregations.IAggregation;
import info.textgrid.utils.export.aggregations.IAggregationEntry;

import java.net.URI;
import java.text.CharacterIterator;
import java.text.StringCharacterIterator;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.google.common.base.Function;
import com.google.common.base.Joiner;
import com.google.common.base.Optional;
import com.google.common.base.Predicate;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableList.Builder;
import com.google.common.collect.Iterables;
import com.google.common.collect.Maps;
import com.ibm.icu.text.Transliterator;

/**
 * An {@link IFilenamePolicy} that is user-customizable by providing a pattern
 * string.
 *
 * <h3>Pattern Syntax</h3>
 *
 * <p>
 * A pattern string is a string containing <em>patterns</em> enclosed in curly
 * braces. Each pattern starts with a <em>variable</em> and is optionally
 * followed by one or more options, each introduced by a vertical bar (
 * <code>|</code>). Please note that all whitespace is significant.
 * </p>
 *
 * As an example, the string
 * <code>{author|fallback|20}-{title|sep=,}.{uri}.{ext}</code> contains the
 * variables <var>author</var> with the options <code>fallback</code> and
 * <code>20</code>, the variable <var>title</var> with the option
 * <code>sep=,</code>, and the variables <var>uri</uri> and <var>ext</var>, each
 * without any option.
 *
 * <h4>Basic Variables</h4> The following basic variables are available in all
 * policies:
 *
 * <table>
 * <thead>
 * <th>Variable</th>
 * <th>Supported Options</th>
 * <th>Description</th>
 * </thead> <tbody>
 * <tr>
 * <td>author</td>
 * <td><code>fallback</code>, <code>sep=</code><var>String</var>,
 * <var>Number</var>, <code>raw</code></td>
 * <td>The object's author. This tries to find the nearest work object in the
 * aggregation tree and extracts its author or authors.
 * <p>
 * If the <code>fallback</code> option is included and the matching work
 * <em>does not</em> include author fields, use all agents regardless of their
 * role instead.
 * </p>
 * </td>
 * </tr>
 *
 * <tr>
 * <td>title</td>
 * <td><code>sep=</code><var>String</var>, <var>Number</var>, <code>raw</code></td>
 * <td>The object's title or titles.</td>
 * </tr>
 *
 * <tr>
 * <td>uri</td>
 * <td>—</td>
 * <td>The object's TextGrid URI. This only includes the scheme-specific part.</td>
 * </tr>
 *
 * <tr>
 * <td>ext</td>
 * <td>—</td>
 * <td>A filename extension that is suitable for the object's MIME type, or
 * <code>dat</code> if none found. This does <em>not</em> include a leading dot.
 * </td>
 * </tr>
 *
 * </tbody>
 * </table>
 *
 * <h4>Options</h4>
 *
 * <table>
 *
 * <tr>
 * <td><var>Number</var></td>
 * <td>
 * If you pass any non-negative non-zero integer number as an option, the
 * expanded value of the variable will be trimmed after at most
 * <var>Number</var> characters. Trimming occurs after all other processing
 * steps for the variable.</td>
 * </tr>
 *
 * <tr>
 * <td><code>raw</code></td>
 * <td>Insert the result of this variable as-is, without character sanitization.
 * <p>
 * If you <em>do not</em> include this option, the result of the metadata-based
 * variables will be transcribed from its original characters to a safe subset
 * of US-ASCII characters in order to be safe from all kinds of encoding and
 * filename issues. This tries to do something sensible with, e.g., umlauts and
 * non-latin scripts.
 * </p>
 * </td>
 * </tr>
 *
 * <tr>
 * <td><code>sep=</code><var>String</var></code></td>
 * <td>If present and the respective metadata field contains multiple values,
 * use all values, joined together with the given separator <var>String</var>.
 * Otherwise, only use the first value.</td>
 * </tr>
 *
 * <tr>
 * <td><code>fallback</code></td>
 * <td>See at the corresponding variable descriptions.</td>
 * </tr>
 * </table>
 *
 */
public class ConfigurableFilenamePolicy implements IFilenamePolicy {

	public class SubPolicySegment extends MetadataSegment implements Segment {

		private ConfigurableFilenamePolicy policy;

		public SubPolicySegment(final ConfigurableFilenamePolicy policy,
				final String... options) {
			super(options);
			this.policy = policy;
			this.raw = true;
		}

		@Override
		public String getContent(final IAggregationEntry entry) {
			return policy.getFilename(entry).toString();
		}

	}

	private final Map<String, ConfigurableFilenamePolicy> subPolicies = Maps
			.newHashMap();

	/**
	 * Adds a sub policy to this policy. If the pattern contains the variable
	 * <var>name</var>, it will be expanded by calling the policy that is
	 * configured here.
	 *
	 * @param name
	 * @param policy
	 * @return
	 */
	public ConfigurableFilenamePolicy addSubPolicy(final String name,
			final ConfigurableFilenamePolicy policy) {
		subPolicies.put(name, policy);
		return this;
	}

	public class ExtensionSegment extends MetadataSegment implements Segment {

		public ExtensionSegment(final String... options) {
			super(options);
			raw = true; // extensions need never be transliterated.
		}

		@Override
		public String getContent(final IAggregationEntry entry) {
			final String format = entry.getMetadata().getGeneric()
					.getProvided().getFormat();
			return FileExtensionMap.getInstance().getFirstExtension(format)
					.or("dat");
		}

	}

	/** A segment represents a part of a filename pattern. */
	protected interface Segment {
		/**
		 * Appends the expansion of this segment for the given entry to the
		 * given builder.
		 */
		public StringBuilder append(final StringBuilder builder,
				IAggregationEntry entry);
	}

	/**
	 * A segment that always expands to a static string, regardless of the entry
	 */
	protected class StaticSegment implements Segment {
		protected final String content;

		/**
		 * Creates a static string segment.
		 *
		 * @param content
		 *            The string this segment will expand to.
		 */
		public StaticSegment(final String content) {
			this.content = content;
		}

		@Override
		public StringBuilder append(final StringBuilder builder,
				final IAggregationEntry entry) {
			return builder.append(content);
		}

	}

	/** Base class for segments that extract metadata from the given argument. */
	protected abstract class MetadataSegment implements Segment {
		/**
		 * Hint to insert the raw content of the metadata field, without any
		 * translation to safe characters.
		 */
		protected boolean raw = false;

		/**
		 * Hint that suggests the maximal number of characters to append for
		 * this segment. Values <= 0 mean no restriction.
		 */
		protected int maxlength = -1;

		/**
		 * Initializes this segment. Calls {@link #handleOption(String)} for
		 * each given option.
		 */
		public MetadataSegment(final String... options) {
			for (final String option : options)
				handleOption(option);
		}

		/**
		 * Parse the given option. Clients may override or extend.
		 *
		 * @param option
		 *            The option. Currently supported strings are
		 *            <code>raw</code> to set the {@link #raw} flag or any
		 *            positive decimal integer to set the {@link #maxlength}
		 *            option.
		 * @return <code>true</code> iff the method handled the option
		 */
		protected boolean handleOption(final String option) {
			if ("raw".equals(option))
				raw = true;
			else if (option.matches("\\d+"))
				maxlength = Integer.parseInt(option);
			else
				return false;
			return true;

		}

		/** Return the (unprocessed) string for this entry. */
		public abstract String getContent(IAggregationEntry entry);

		/**
		 * {@inheritDoc}
		 *
		 * The {@link MetadataSegment} implementation calls
		 * {@link #getContent(IAggregationEntry)} and postprocesses the result
		 * according to the {@link #raw} and {@link #maxlength} fields before
		 * appending.
		 */
		@Override
		public StringBuilder append(final StringBuilder builder,
				final IAggregationEntry entry) {
			String content = getContent(entry);
			if (!raw)
				content = translate(content);
			if (maxlength > 0)
				content = content.substring(0, maxlength - 1);
			return builder.append(content);
		}

	}

	/**
	 * Formats an object's TextGrid URI.
	 */
	public class URISegment extends MetadataSegment implements Segment {

		public URISegment(final String... options) {
			super(options);
		}

		@Override
		public String getContent(final IAggregationEntry entry) {
			return entry.getMetadata().getGeneric().getGenerated()
					.getTextgridUri().getValue();
		}

	}

	/**
	 * Extracts the title.
	 *
	 * If the additional option sep=<string> is given and the given object has
	 * multiple
	 *
	 */
	public abstract class ListSegment extends MetadataSegment {

		protected Optional<String> separator = Optional.absent();

		public ListSegment(final String... options) {
			super(options);
		}

		private final Pattern SEP = Pattern.compile("^sep=(.+)$");

		@Override
		protected boolean handleOption(final String option) {
			if (super.handleOption(option))
				return true;
			else {
				final Matcher matcher = SEP.matcher(option);
				if (matcher.matches()) {
					separator = Optional.of(matcher.group(1));
					return true;
				}
				return false;
			}
		}

		public abstract Iterable<String> getContentItems(
				final IAggregationEntry entry);

		@Override
		public String getContent(final IAggregationEntry entry) {
			final Iterable<String> items = getContentItems(entry);
			if (separator.isPresent())
				return Joiner.on(separator.get()).join(items);
			else {
				final Iterator<String> iter = items.iterator();
				if (iter.hasNext())
					return iter.next();
				else
					return "";
			}
		}
	}

	public class TitleSegment extends ListSegment {

		@Override
		public Iterable<String> getContentItems(final IAggregationEntry entry) {
			return entry.getMetadata().getGeneric().getProvided().getTitle();
		}

		public TitleSegment(final String... options) {
			super(options);
		}

	}

	public class AuthorSegment extends ListSegment {

		private final Function<AgentType, String> AGENT_NAME = new Function<AgentType, String>() {

			@Override
			public String apply(final AgentType input) {
				return input.getValue();
			}
		};

		public AuthorSegment(final String... options) {
			super(options);
		}

		private boolean fallback = false;

		@Override
		protected boolean handleOption(final String option) {
			if (super.handleOption(option))
				return true;
			else if ("fallback".equals(option)) {
				this.fallback = true;
				return true;
			}
			return false;
		}

		@Override
		public Iterable<String> getContentItems(final IAggregationEntry entry) {
			if (entry.getMetadata().getWork() != null) {
				final List<AgentType> agents = entry.getMetadata().getWork()
						.getAgent();
				final Iterable<AgentType> authors = Iterables.filter(agents,
						new Predicate<AgentType>() {

							@Override
							public boolean apply(final AgentType agent) {
								return AgentRoleType.AUTHOR.equals(agent
										.getRole());
							}
						});
				if (authors.iterator().hasNext()) {
					return Iterables.transform(authors, AGENT_NAME);
				} else if (fallback)
					return Iterables.transform(agents, AGENT_NAME);
			} else if (entry instanceof IAggregation
					&& ((IAggregation) entry).getWork().isPresent()) {
				return getContentItems(((IAggregation) entry).getWork().get());
			} else if (entry.getParent().isPresent())
				return getContentItems(entry.getParent().get());
			return null;
		}

	}

	private final Transliterator transliterator;
	private ImmutableList<Segment> segments;
	private boolean isParentPolicy;

	public ConfigurableFilenamePolicy(final String pattern) {
		this(pattern, false);
	}

	public ConfigurableFilenamePolicy(final String pattern,
			final boolean isParentPolicy) {
		this.isParentPolicy = isParentPolicy;

		this.transliterator = DefaultFilenamePolicy.getTransliterator();

		StringBuilder buffer = new StringBuilder();
		final Builder<Segment> segments = ImmutableList.builder();
		boolean escape = false;

		final StringCharacterIterator iter = new StringCharacterIterator(
				pattern);

		for (char c = iter.first(); c != CharacterIterator.DONE; c = iter
				.next())
			if (escape) {
				buffer.append(c);
				escape = false;
			} else if (c == '\\')
				escape = true;
			else if (c == '{') {
				final String s = buffer.toString();
				if (!s.isEmpty()) {
					segments.add(new StaticSegment(s));
					buffer = new StringBuilder();
				}
				segments.add(parseVariable(iter));
			} else
				buffer.append(c);
		this.segments = segments.build();
	}

	private Segment parseVariable(final StringCharacterIterator iter) {
		if (iter.current() == '{')
			iter.next();
		final StringBuilder buf = new StringBuilder();
		for (char c = iter.current(); c != '}' && c != CharacterIterator.DONE; c = iter
				.next())
			buf.append(c);
		final String[] split = buf.toString().split(",");
		final String var = split[0];
		final String[] options = Arrays.copyOfRange(split, 1, split.length);

		if (subPolicies.containsKey(var))
			return new SubPolicySegment(subPolicies.get(var), options);
		else if (("uri").equals(var))
			return new URISegment(options);
		else if (("title".equals(var)))
			return new TitleSegment(options);
		else if (("ext".equals(var)))
			return new ExtensionSegment(options);
		return null;
	}

	@Override
	public String translate(final String source) {
		return transliterator.transliterate(source);
	}

	@Override
	public String getFilename(final ObjectType object, final boolean asParent) {
		return null;
	}

	/**
	 * @deprecated Use {@link #getFilename(IAggregationEntry)} instead
	 */
	@Deprecated
	@Override
	public URI getFilename(final IAggregationEntry entry, final boolean asParent) {
		return getFilename(entry);
	}

	@Override
	public URI getFilename(final IAggregationEntry entry) {
		final StringBuilder builder = new StringBuilder();
		for (final Segment segment : segments)
			segment.append(builder, entry);
		return URI.create(builder.toString());
	}

	@Override
	public Optional<URI> getBase(final IAggregationEntry entry) {
		// TODO Auto-generated method stub
		return null;
	}
}
